#include "streamplayer.h"
#include "rtspplayer.h"
#include "debug.h"

#include "liveMedia/liveMedia.hh"
#include "BasicUsageEnvironment.hh"
#include "GroupsockHelper.hh"



class cRTSPHandler;

#define TIMEDEBUG
#ifdef TIMEDEBUG
#include <sys/time.h>
#include <time.h>
class hstimer {
	timeval t0,t1;
	int t;
	int max;
	int sum;
public:
	hstimer() { Start(); clear(); t1=t0; }
	void Start() { gettimeofday(&t0,NULL); }
	void Stop() { gettimeofday(&t1,NULL); }
	int time() { return (t1.tv_sec-t0.tv_sec)*1000000 + t1.tv_usec - t0.tv_usec; }
	void check() { int t=time(); if (t>max) max=t; sum+=t; }
	void clear() { max=0; sum=0; }
	int getmax() { return max; }
	int getsum() { return sum; }
};
hstimer gptime;
#endif


//
// cVDRSink
//

class cVDRSink: public MediaSink {
	// liveMedia sink, accepts incoming data for one (video or audio) data stream.
	// Data is repacked to PES and forwarded to cRTSPHandler::Put

public:
  cVDRSink(UsageEnvironment& env, cRTSPHandler *Handler, int StreamID, int InitialFrameSize);
  virtual ~cVDRSink();

protected:
	struct SimplePESHeader {
		unsigned char packet_start_code_prefix[3];
		unsigned char stream_id;
		unsigned char PES_packet_length_0,PES_packet_length_1;
		unsigned char flags1;
		unsigned char flags2;
		unsigned char PES_header_data_length;
		unsigned char PTS[5];
	};
	static const int SimplePESHeaderSize = sizeof(SimplePESHeader);
	static const int PESMaxDataSize = 65535 - SimplePESHeaderSize + 6;

  unsigned char *buffer;
  int bufferSize;
  unsigned int maxFrameSize;
  int stream_id;
  cRTSPHandler *handler;

  int WritePESHeader(SimplePESHeader *header, int size, timeval PTS);

  static void staticAfterGettingFrame(void* clientData, unsigned frameSize,
				unsigned numTruncatedBytes,
				struct timeval presentationTime,
				unsigned durationInMicroseconds);
  virtual void afterGettingFrame(unsigned frameSize,
				unsigned numTruncatedBytes,
				struct timeval presentationTime,
				unsigned durationInMicroseconds);

private:
  virtual Boolean continuePlaying();

};

//
// cVDRSubsession
//

class cVDRSubsession {
	// Bundles liveMedia subsession information.
	// Calls cRTSPHandler::afterPlaying() after last subsession has ended.

	MediaSubsession *Subsession;
	cRTSPHandler *handler;
public:
	cVDRSubsession(MediaSubsession *subsession, cRTSPHandler *Handler);
	MediaSubsession* subsession() { return Subsession; }

	Boolean attachSink(MediaSink *sink);

protected:
  static void staticAfterPlaying(void* clientData);
  void afterPlaying();

};

//
// cRTSPHandler
//

class cRTSPHandler : public cThread {
	// Main handler thread for liveMedia
	// liveMedia is not thread safe, and uses a cooperative task scheduler,
	// running in this thread only.


public:
	cRTSPHandler(char *Url);

	virtual ~cRTSPHandler();

	void Stop();

protected:
	bool Active;
	bool shutdown;
	cRingBufferLinear *buffer;
	char *url;

	cMutex Scheduler;
	cMutex SchedulerControl;
	char ExternalEvent; // protected by TaskSchedulerControl

	void LockTaskScheduler() {
		SchedulerControl.Lock();
		ExternalEvent=1;
		Scheduler.Lock(); // wait for scheduler to react
		ExternalEvent=0;
		SchedulerControl.Unlock();
	}
	void UnlockTaskScheduler() {
		Scheduler.Unlock();
	}


	virtual void Action(void);

	TaskScheduler* scheduler;
	UsageEnvironment* env;
	MediaSession *session;
	RTSPClient *client;
	timeval lastPTS,startPTS;

	cVDRSubsession *audio,*video;

	bool CreateSession();
	void DestroySession();

	void afterPlaying();

public:
	float GetPCR();
	void SetPTS(timeval PTS);

	int Put(const uchar *Data, int Count) {
		// Put PES data. Called from cVDRSink.
		if (!buffer) return 0;
		return buffer->Put(Data,Count);
	}

	uchar *Get(int &Count) {
		// Get buffered data. Called from other thread.
		if (!buffer) { Count=0; return NULL; }
		uchar *data=buffer->Get(Count);
		// returning NULL means we just waited for data,
		// so data may now be available...
		if (!data) data=buffer->Get(Count);
		return data;
	}

	void Del(int Count) {
		// Delete buffered data. Called from other thread.
		if (buffer) buffer->Del(Count);
	}

	int Available() {
		if (!buffer) return 0;
		return buffer->Available();
	}


	void PlayMediaSession(float start=0.0f, float end=-1.0f, float scale=1.0f);
	void PauseMediaSession();


	friend class cVDRSubsession;
};



//
// cStreamPlayerRTSP
//

class cStreamPlayerRTSP : public cStreamPlayer {
	// Stream player playback task
	// Polls buffered data from RTSP task and forwards it to VDR

protected:
	cRTSPHandler rtspHandler;

	bool pause;
	int MinCollect;
    char StatusString[80];

	virtual void Action(void);

public:
	cStreamPlayerRTSP(char *url);
	virtual ~cStreamPlayerRTSP();

	virtual void Pause();
    virtual void SkipFwd();
    virtual void SkipBwd();

	virtual const char* GetStatus();
};









//
// cVDRSink
//


cVDRSink::cVDRSink(UsageEnvironment& env, cRTSPHandler *Handler, int StreamID, int InitialFrameSize)
	: MediaSink(env) {

	handler=Handler;

	maxFrameSize=0;
	bufferSize=InitialFrameSize;
	buffer=(unsigned char*)malloc(bufferSize);

	stream_id=StreamID;
}

cVDRSink::~cVDRSink() {
	_dprintf("Sink: maxFrameSize=%i\n",maxFrameSize);
	free(buffer);
}

int cVDRSink::WritePESHeader(SimplePESHeader *header, int frameSize, timeval PTS) {

	header->packet_start_code_prefix[0]=0;
	header->packet_start_code_prefix[1]=0;
	header->packet_start_code_prefix[2]=1;
	header->stream_id=stream_id;

	if (frameSize>PESMaxDataSize) {
		frameSize=PESMaxDataSize;
	}
	int frameHeaderSize=frameSize+SimplePESHeaderSize-6;

	((SimplePESHeader*)buffer)->PES_packet_length_0=frameHeaderSize >> 8;
	((SimplePESHeader*)buffer)->PES_packet_length_1=frameHeaderSize & 0xff;

	header->flags1=0x80;
	header->flags2=0x80;
	header->PES_header_data_length=5;

	// Calculate time in 90Khz steps, but as 33-bit number
	bool HighBit=((PTS.tv_sec*45000 + (PTS.tv_usec*9)/200)& 0x80000000) != 0;
	unsigned int LowBits=PTS.tv_sec*90000 + (PTS.tv_usec*9)/100;

	header->PTS[0]=0x20 | (HighBit?0x08:0x00) | ((LowBits>>29)&0x06) | 0x01;
	header->PTS[1]=(LowBits>>22)&0xff;
	header->PTS[2]=((LowBits>>14)&0xff) | 0x01;
	header->PTS[3]=(LowBits>>7)&0xff;
	header->PTS[4]=((LowBits<<1)&0xfe) | 0x01;

	return frameSize;
}


void cVDRSink::staticAfterGettingFrame(void* clientData, unsigned frameSize, unsigned numTruncatedBytes,
	                                   struct timeval presentationTime,	unsigned durationInMicroseconds) {

	((cVDRSink*)clientData)->afterGettingFrame(frameSize, numTruncatedBytes, presentationTime, durationInMicroseconds);
}

void cVDRSink::afterGettingFrame(unsigned frameSize, unsigned numTruncatedBytes,
	                             struct timeval presentationTime, unsigned durationInMicroseconds) {

	unsigned char *myBuffer=buffer;
	handler->SetPTS(presentationTime);
	if (frameSize>maxFrameSize) maxFrameSize=frameSize;

	while (frameSize>0) {
		int taken=WritePESHeader((SimplePESHeader*)myBuffer,frameSize,presentationTime);
		int accepted=handler->Put(myBuffer,taken+SimplePESHeaderSize);
		if (accepted!=taken+SimplePESHeaderSize) {
			_dprintf("Buffer overrun occured, %i bytes lost\n",taken+SimplePESHeaderSize-accepted);
			break;
		}

		myBuffer+=taken;
		frameSize-=taken;
	}
	if (numTruncatedBytes>0) {
		// increase buffer size
		int newBufferSize=(bufferSize+numTruncatedBytes+0xfff)&~0xfff;
		free(buffer);
		buffer=(unsigned char*)malloc(newBufferSize);
		bufferSize=newBufferSize;
		_dprintf("Increased frame buffer to %ik\n",bufferSize/1024);
	}

	continuePlaying();
}


Boolean cVDRSink::continuePlaying() {
	if (fSource == NULL) return False;
	fSource->getNextFrame(buffer+SimplePESHeaderSize, bufferSize-SimplePESHeaderSize, staticAfterGettingFrame, this, onSourceClosure, this);
	return True;
}


//
// cVDRSubsession
//

cVDRSubsession::cVDRSubsession(MediaSubsession *subsession, cRTSPHandler *Handler) {
	Subsession=subsession;
	handler=Handler;
}

Boolean cVDRSubsession::attachSink(MediaSink *sink) {
	Subsession->sink=sink;
	return sink->startPlaying(*(Subsession->readSource()),staticAfterPlaying,this);
}

void cVDRSubsession::staticAfterPlaying(void* clientData) { ((cVDRSubsession*)clientData)->afterPlaying(); }
void cVDRSubsession::afterPlaying() {
	Medium::close(Subsession->sink);
  	Subsession->sink = NULL;

  	MediaSession& session = Subsession->parentSession();
  	MediaSubsessionIterator iter(session);
  	while ((Subsession = iter.next()) != NULL) {
    	if (Subsession->sink != NULL) return; // this subsession is still active
  	}

  	// All subsessions have now been closed
	handler->afterPlaying();
}



//
// cRTSPHandler
//

cRTSPHandler::cRTSPHandler(char *Url) {
	scheduler = BasicTaskScheduler::createNew();
	env = BasicUsageEnvironment::createNew(*scheduler);
	asprintf(&url,"rtsp:%s",Url);
	session=NULL;
	client=NULL;
	audio=NULL;
	video=NULL;
	Active=false;
	shutdown=false;
	ExternalEvent=0;
	buffer=NULL;
	startPTS.tv_sec=0;
	startPTS.tv_usec=0;
	lastPTS.tv_sec=0;
	lastPTS.tv_usec=0;
}


cRTSPHandler::~cRTSPHandler() {
	Stop();

	if (buffer) delete buffer;
	if (Active) _dprintf("WARNING: destroyed task is active!\n");
	if (env) env->reclaim();
	if (scheduler) delete scheduler;
	if (url) free(url);
}

void cRTSPHandler::Stop() {
	if (!Active) return;
	shutdown=true;
	LockTaskScheduler(); // make sure task scheduler exits
	UnlockTaskScheduler();
	Cancel(2);
}

bool cRTSPHandler::CreateSession() {
	client = RTSPClient::createNew(*env, 1, "Stream Player");
	if (!client) {
		_dprintf("create client failed\n");
		return false;
	}

	_dprintf("create client ok\n");

	/*
	char* optionsResponse = rtspClient->sendOptionsCmd(url);
	delete[] optionsResponse;
	*/

	char* sdpDescription = client->describeURL(url);
	if (!sdpDescription) {
		_dprintf("getSDPDescriptionFromURL \"%s\" failed\n",url);
		return false;
	}

	_dprintf("describeURL ok\n");

	session = MediaSession::createNew(*env, sdpDescription);
	delete[] sdpDescription;

	_dprintf("create session ok\n");

	if (!session->hasSubsessions()) {
		_dprintf("No subsessions\n");
		return false;
	}

	MediaSubsessionIterator iter(*session);
	MediaSubsession *subsession;

	while ((subsession = iter.next()) != NULL) {
		if (0!=strcmp(subsession->mediumName(),"video")) continue;

		if (   0==strcmp(subsession->codecName(),"MPV")) {
			video=new cVDRSubsession(subsession,this);
			break;
		}
	}

	iter.reset();
	while ((subsession = iter.next()) != NULL) {
		if (0!=strcmp(subsession->mediumName(),"audio")) continue;

		if (   0==strcmp(subsession->codecName(),"MPA")
		    || 0==strcmp(subsession->codecName(),"MPA-ROBUST")
		   ) {
			audio=new cVDRSubsession(subsession,this);
			break;
		}
	}

	_dprintf("scan video audio ok\n");

	if (!video || !audio) {
		_dprintf("No video or audio subsession found\n");
		return false;
	}

	if (!video->subsession()->initiate()) {
		_dprintf("subsession initiate failed: %s.\n",env->getResultMsg());
		return false;
	}
	if (!audio->subsession()->initiate()) {
		_dprintf("subsession initiate failed: %s.\n",env->getResultMsg());
		return false;
	}
	_dprintf("Port number: video %i audio %i\n",video->subsession()->clientPortNum(),audio->subsession()->clientPortNum());

	_dprintf("initiate ok\n");


	if( audio->subsession()->rtpSource() ) {
		/* Increase the buffer size */
		int fd = audio->subsession()->rtpSource()->RTPgs()->socketNum();
		increaseReceiveBufferTo(*env, fd, 100000);
	}

	if( video->subsession()->rtpSource() ) {
		/* Increase the buffer size */
		int fd = video->subsession()->rtpSource()->RTPgs()->socketNum();
		increaseReceiveBufferTo(*env, fd, 2000000);
	}


	const Boolean UseTCP=False;

	if (!client->setupMediaSubsession(*audio->subsession(),False,UseTCP)) {
		_dprintf("setup audio subsession failed: %s.\n",env->getResultMsg());
		return false;
	}
	if (!client->setupMediaSubsession(*video->subsession(),False,UseTCP)) {
		_dprintf("setup video subsession failed: %s.\n",env->getResultMsg());
		return false;
	}

	_dprintf("setupMediaSubsession ok\n");

	if (!audio->attachSink(new cVDRSink(*env,this,0xC0,0x1000))) {
		_dprintf("attach audio sink failed: %s.\n",env->getResultMsg());
	}
	if (!video->attachSink(new cVDRSink(*env,this,0xE0,0x10000))) {
		_dprintf("attach video sink failed: %s.\n",env->getResultMsg());
	}

	_dprintf("attach sink ok\n");

	if (!client->playMediaSession(*session)) {
		_dprintf("playMediaSession failed: %s.\n",env->getResultMsg());
		return false;
	}

	return true;
}

void cRTSPHandler::DestroySession() {
	_dprintf("DestroySession\n");
	if (session) {
		MediaSubsessionIterator iter(*session);
		MediaSubsession* subsession;
		while ((subsession = iter.next()) != NULL) {
			if (subsession->sink) {
				_dprintf("Closing a subsession sink\n");
				Medium::close(subsession->sink);
				subsession->sink = NULL;
			}
		}
		_dprintf("Closing VDRSubsessions\n");
		if (audio) delete audio;
		if (video) delete video;

		_dprintf("Calling Teardown\n");
		client->teardownMediaSession(*session);

		_dprintf("Closing session\n");
		Medium::close(session);
		session=NULL;
	}
	if (client) {
		_dprintf("Closing client\n");
		Medium::close(client);
		client=NULL;
	}
}


void cRTSPHandler::afterPlaying() {
	shutdown=true;
	ExternalEvent=1; // from within thread, we cannot lock TaskSchedulerControl...
}

void cRTSPHandler::Action(void) {
	Active=true;
	shutdown=false;

	if (!buffer) buffer = new cRingBufferLinear(cPluginStreamplayer::Plugin->BufferSize*1024);
	buffer->SetTimeouts(5,50);

	if (!CreateSession()) {
		Active=false;
		shutdown=true;
		return;
	}

	while (!shutdown) {
		if (ExternalEvent!=0) {
			usleep(3); // Nothing to do.
		} else {
			Scheduler.Lock();
			env->taskScheduler().doEventLoop((char*)&ExternalEvent);
			Scheduler.Unlock();
		}
	}

	_dprintf("taskScheduler terminated.\n");

    DestroySession();

	Active=false;
	shutdown=false;
}

float cRTSPHandler::GetPCR() {
	return (float)lastPTS.tv_sec - (float)startPTS.tv_sec
	       + ((float)lastPTS.tv_usec - (float)startPTS.tv_usec) / 1000000.0f;
}

void cRTSPHandler::SetPTS(timeval PTS) {
	lastPTS=PTS;
	if (   (startPTS.tv_sec == 0 && startPTS.tv_usec == 0)
		|| (startPTS.tv_sec > lastPTS.tv_sec)
		|| (startPTS.tv_sec == lastPTS.tv_sec && startPTS.tv_usec > lastPTS.tv_usec)) {

		startPTS=lastPTS;
	}

}


void cRTSPHandler::PlayMediaSession(float start, float end, float scale) {
	LockTaskScheduler();
	client->playMediaSession(*session,start,end,scale);
	UnlockTaskScheduler();
}
void cRTSPHandler::PauseMediaSession() {
	LockTaskScheduler();
	client->pauseMediaSession(*session);
	UnlockTaskScheduler();
}





//
// cStreamPlayerRTSP
//


cStreamPlayerRTSP::cStreamPlayerRTSP(char *url)
	: rtspHandler(url)	{
	pause=false;
}
cStreamPlayer* CreateRTSPPlayer(char *url) { return new cStreamPlayerRTSP(url); }


cStreamPlayerRTSP::~cStreamPlayerRTSP() {
	rtspHandler.Stop();

	// Kill thread before deepest nested destructor finishes
	// Not good style, but not to avoid.
	Detach();
	Stop();

	if (active) _dprintf("WARNING: destroyed task is active!\n");
}

const char* cStreamPlayerRTSP::GetStatus() {

	int available=rtspHandler.Available();

	snprintf(StatusString,sizeof StatusString,"(buf %ik)",available/1024);
	return StatusString;
}

void cStreamPlayerRTSP::Pause() {
	if (pause) {
		rtspHandler.PlayMediaSession(-1.0);
		pause=false;
	} else {
		rtspHandler.PauseMediaSession();
		pause=true;
	}
}

void cStreamPlayerRTSP::SkipFwd() {
	if (!pause) rtspHandler.PauseMediaSession();
	rtspHandler.PlayMediaSession(rtspHandler.GetPCR()+60);
	pause=false;
}

void cStreamPlayerRTSP::SkipBwd() {
	if (!pause) rtspHandler.PauseMediaSession();
	float pts=rtspHandler.GetPCR()-60;
	if (pts<0) pts=0;
	rtspHandler.PlayMediaSession(pts);
	pause=false;
}

void cStreamPlayerRTSP::Action(void) {
	hstimer t;

	DevicePlay();

	rtspHandler.Start();
	MinCollect=cPluginStreamplayer::Plugin->BufferPreFill*1024;

	while(active && !closedown) {
		uchar *pPlayData;
		int nPlayData;

		t.Start();

		// Get data from the RTSP task
		pPlayData=rtspHandler.Get(nPlayData);

		t.Stop();
		if (t.time()>40000) _dprintf("rtspHandler.Get() %ius\n",t.time());

		// If play buffer is filled, transmit to player.
		if (pPlayData!=NULL && nPlayData>MinCollect) {

			// limit amount of data:
			if (nPlayData>128*1024) nPlayData=128*1024;

			t.Start();
			int r=PlayPes(pPlayData,nPlayData);
			t.Stop();
			if (t.time()>40000) _dprintf("PlayPes() %ius\n",t.time());


			if (r>0) rtspHandler.Del(r);
			MinCollect=0;
		} else {
			// Nothing to do. Sleep.
			t.Start();
			usleep(3);
			t.Stop();
			if (t.time()>40000) _dprintf("usleep() %ius\n",t.time());
		}
	}

	rtspHandler.Stop();

	active=false;
	closedown=false;
}


